#include <stdio.h>
#include <limits.h>
#include <assert.h>
#include <goatvr.h>
#include "app.h"
#include "opengl.h"
#include "sdr.h"
#include "texture.h"
#include "mesh.h"
#include "meshgen.h"
#include "scene.h"
#include "metascene.h"
#include "datamap.h"
#include "ui.h"
#include "opt.h"
#include "post.h"
#include "renderer.h"
#include "rtarg.h"
#include "avatar.h"
#include "vrinput.h"
#include "exman.h"
#include "blob_exhibit.h"
#include "dbg_gui.h"
#include "geomdraw.h"
#include "ui_exhibit.h"

#define NEAR_CLIP	5.0
#define FAR_CLIP	10000.0

static void draw_scene();
static void toggle_flight();
static void calc_framerate();
static Ray calc_pick_ray(int x, int y);

long time_msec;
int win_width, win_height;
int vp_width, vp_height;
float win_aspect;
bool fb_srgb;
bool opt_gear_wireframe;

TextureSet texman;
SceneSet sceneman;

int fpexcept_enabled;

unsigned int dbg_key_pending;

static Avatar avatar;

static float cam_dist = 0.0;
static float floor_y;	// last floor height
static float user_eye_height = 165;

static float walk_speed = 300.0f;
static float mouse_speed = 0.5f;
static bool show_walk_mesh, noclip = false;

static bool have_headtracking, have_handtracking, should_swap;

static int prev_mx, prev_my;
static bool bnstate[8];
static bool keystate[256];
static bool gpad_bnstate[64];
static Vec2 joy_move, joy_look;
static float joy_deadzone = 0.01;

static float framerate;

static Mat4 view_matrix, mouse_view_matrix, proj_matrix;
static MetaScene *mscn;
static unsigned int sdr_post_gamma;

static long prev_msec;

static ExhibitManager *exman;
static bool show_blobs;

ExSelection exsel_active, exsel_hover;
ExSelection exsel_grab_left, exsel_grab_right;
#define exsel_grab_mouse exsel_grab_right
static ExhibitSlot exslot_left, exslot_right;
#define exslot_mouse exslot_right

static bool pointing;

static Renderer *rend;
static RenderTarget *goatvr_rtarg;

static Ray last_pick_ray;

static bool post_scene_init_pending = true;


bool app_init(int argc, char **argv)
{
	set_log_file("demo.log");

	char *env = getenv("FPEXCEPT");
	if(env && atoi(env)) {
		info_log("enabling floating point exceptions\n");
		fpexcept_enabled = 1;
		enable_fpexcept();
	}

	if(init_opengl() == -1) {
		return false;
	}

	if(!init_options(argc, argv, "demo.conf")) {
		return false;
	}
	app_resize(opt.width, opt.height);
	app_fullscreen(opt.fullscreen);

	if(opt.vr) {
		if(goatvr_init() == -1) {
			return false;
		}
		goatvr_set_origin_mode(GOATVR_HEAD);
		goatvr_set_units_scale(100.0f);

		goatvr_startvr();
		should_swap = goatvr_should_swap() != 0;
		user_eye_height = goatvr_get_eye_height();
		have_headtracking = goatvr_have_headtracking();
		have_handtracking = goatvr_have_handtracking();

		goatvr_recenter();

		goatvr_rtarg = new RenderTarget;
	}

	if(fb_srgb) {
		int srgb_capable;
		glGetIntegerv(GL_FRAMEBUFFER_SRGB_CAPABLE_EXT, &srgb_capable);
		printf("Framebuffer %s sRGB-capable\n", srgb_capable ? "is" : "is not");
		if(srgb_capable) {
			glEnable(GL_FRAMEBUFFER_SRGB);
		} else {
			fb_srgb = 0;
		}
	}

	glEnable(GL_MULTISAMPLE);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	glEnable(GL_NORMALIZE);

	if(!init_debug_gui()) {
		return false;
	}

	Mesh::use_custom_sdr_attr = false;

	float ambient[] = {0.0, 0.0, 0.0, 0.0};
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient);

	init_audio();

	if(!init_vrhands()) {
		return false;
	}

	mscn = new MetaScene;
	if(!mscn->load(opt.scenefile ? opt.scenefile : "data/museum.scene")) {
		return false;
	}

	avatar.pos = mscn->start_pos;
	Vec3 dir = rotate(Vec3(0, 0, 1), mscn->start_rot);
	dir.y = 0;
	avatar.body_rot = rad_to_deg(acos(dot(dir, Vec3(0, 0, 1))));

	exman = new ExhibitManager;
	// exhibits are loaded in post_scene_init, because they need access to the scene graph

	if(!exui_init()) {
		error_log("failed to initialize exhibit ui system\n");
		return false;
	}
	exui_setnode(&exslot_left.node);

	if(!fb_srgb) {
		sdr_post_gamma = create_program_load("sdr/post_gamma.v.glsl", "sdr/post_gamma.p.glsl");
	}

	rend = new Renderer;
	if(!rend->init()) {
		return false;
	}
	if(opt.reflect) {
		rend->ropt |= RENDER_MIRRORS;
	} else {
		rend->ropt &= ~RENDER_MIRRORS;
	}
	rend->set_scene(mscn);

	glUseProgram(0);

	if(opt.vr || opt.fullscreen) {
		app_grab_mouse(true);
	}

	if(mscn->music && opt.music) {
		mscn->music->play(AUDIO_PLAYMODE_LOOP);
	}
	return true;
}

// post_scene_init is called after the scene has completed loading
static void post_scene_init()
{
	mscn->update(0);	// update once to calculate node matrices

	int num_mir = mscn->calc_mirror_planes();
	info_log("found %d mirror planes\n", num_mir);

	exman->load(mscn, "data/exhibits");
}

void app_cleanup()
{
	if(mscn->music) {
		mscn->music->stop();
	}
	destroy_audio();

	app_grab_mouse(false);
	if(opt.vr) {
		delete goatvr_rtarg;
		goatvr_shutdown();
	}
	destroy_vrhands();

	delete rend;

	exui_shutdown();

	/* this must be destroyed before the scene graph to detach exhibit nodes
	 * before the scene tries to delete them recursively
	 */
	delete exman;

	texman.clear();
	sceneman.clear();

	cleanup_debug_gui();
}

static bool constrain_walk_mesh(const Vec3 &v, Vec3 *newv)
{
	Mesh *wm = mscn->walk_mesh;
	if(!wm) {
		*newv = v;
		return true;
	}

	Ray downray = Ray(v, Vec3(0, -1, 0));
	HitPoint hit;
	if(mscn->walk_mesh->intersect(downray, &hit)) {
		*newv = hit.pos;
		newv->y += user_eye_height;
		return true;
	}
	return false;
}

static void update(float dt)
{
	texman.update();
	sceneman.update();

	if(post_scene_init_pending && !sceneman.pending()) {
		post_scene_init_pending = false;
		post_scene_init();
	}

	mscn->update(dt);
	exman->update(dt);
	exui_update(dt);

	// use goatvr sticks for joystick input
	int num_vr_sticks = goatvr_num_sticks();
	if(num_vr_sticks > 0) {
		float p[2];
		goatvr_stick_pos(0, p);
		joy_move.x = p[0];
		joy_move.y = -p[1];
	}
	if(num_vr_sticks > 1) {
		float p[2];
		goatvr_stick_pos(1, p);
		joy_look.x = p[0];
	}


	float speed = walk_speed * dt;
	Vec3 dir;

	// joystick
	float jdeadsq = joy_deadzone * joy_deadzone;
	float jmove_lensq = length_sq(joy_move);
	float jlook_lensq = length_sq(joy_look);

	if(jmove_lensq > jdeadsq) {
		float len = sqrt(jmove_lensq);
		jmove_lensq -= jdeadsq;

		float mag = len * len;
		dir.x += mag * joy_move.x / len * 2.0 * speed;
		dir.z += mag * joy_move.y / len * 2.0 * speed;
	}
	if(jlook_lensq > jdeadsq) {
		float len = sqrt(jlook_lensq);
		jlook_lensq -= jdeadsq;

		float mag = len * len;
		avatar.body_rot += mag * joy_look.x / len * 200.0 * dt;
		avatar.head_alt += mag * joy_look.y / len * 100.0 * dt;
		if(avatar.head_alt < -90.0f) avatar.head_alt = -90.0f;
		if(avatar.head_alt > 90.0f) avatar.head_alt = 90.0f;
	}

	// keyboard move
	if(keystate[(int)'w']) {
		dir.z -= speed;
	}
	if(keystate[(int)'s']) {
		dir.z += speed;
	}
	if(keystate[(int)'d']) {
		dir.x += speed;
	}
	if(keystate[(int)'a']) {
		dir.x -= speed;
	}
	if(keystate[(int)'q'] || gpad_bnstate[GPAD_UP]) {
		avatar.pos.y += speed;
	}
	if(keystate[(int)'z'] || gpad_bnstate[GPAD_DOWN]) {
		avatar.pos.y -= speed;
	}

	float theta = M_PI * avatar.body_rot / 180.0f;
	Vec3 newpos;
	newpos.x = avatar.pos.x + cos(theta) * dir.x - sin(theta) * dir.z;
	newpos.y = avatar.pos.y;
	newpos.z = avatar.pos.z + sin(theta) * dir.x + cos(theta) * dir.z;

	if(noclip) {
		avatar.pos = newpos;
	} else {
		if(!constrain_walk_mesh(newpos, &avatar.pos)) {
			float dtheta = M_PI / 32.0;
			float theta = dtheta;
			Vec2 dir2d = newpos.xz() - avatar.pos.xz();

			for(int i=0; i<16; i++) {
				Vec2 dvec = rotate(dir2d, theta);
				Vec3 pos = avatar.pos + Vec3(dvec.x, 0, dvec.y);
				if(constrain_walk_mesh(pos, &avatar.pos)) {
					break;
				}
				dvec = rotate(dir2d, -theta);
				pos = avatar.pos + Vec3(dvec.x, 0, dvec.y);
				if(constrain_walk_mesh(pos, &avatar.pos)) {
					break;
				}
				theta += dtheta;
			}
		}
		floor_y = avatar.pos.y - user_eye_height;
	}

	if(have_headtracking) {
		Quat qhead;
		goatvr_head_orientation(&qhead.x);
		avatar.tracked_head_rotation(qhead);
	}

	// TODO move to the avatar system
	// calculate mouselook view matrix
	mouse_view_matrix = Mat4::identity;
	mouse_view_matrix.pre_translate(0, 0, -cam_dist);
	if(!have_headtracking) {
		mouse_view_matrix.pre_rotate_x(deg_to_rad(avatar.head_alt));
	}
	mouse_view_matrix.pre_rotate_y(deg_to_rad(avatar.body_rot));
	mouse_view_matrix.pre_translate(-avatar.pos.x, -avatar.pos.y, -avatar.pos.z);

	// update hand-tracking
	if(have_handtracking) {
		update_vrhands(&avatar);

		ExSelection *exsel_grab[] = { &exsel_grab_left, &exsel_grab_right };
		ExhibitSlot *exslot[] = { &exslot_left, &exslot_right };

		for(int i=0; i<2; i++) {
			if(vrhand[i].valid) {
				exslot[i]->node.set_position(vrhand[i].pos);
				exslot[i]->node.set_rotation(vrhand[i].rot * exslot[i]->rotation);

				bool act_grab = goatvr_action(i, GOATVR_ACTION_GRAB) != 0;

				ExSelection sel;
				sel = exman->select(Sphere(vrhand[i].pos, 10));

				if(!*exsel_grab[i]) {
					// we don't have an exhibit grabbed
					if(act_grab) {
						// grab an exhibit
						*exsel_grab[i] = sel;
						exslot[i]->rotation = sel.ex->node->get_rotation();
						debug_log("Grabbing with rot: %f %f %f %f\n", exslot[i]->rotation.x,
								exslot[i]->rotation.y, exslot[i]->rotation.z, exslot[i]->rotation.w);
						exslot[i]->attach_exhibit(sel.ex, EXSLOT_ATTACH_TRANSIENT);
						if(exsel_active) {
							exsel_active = ExSelection::null;	// cancel active on grab
						}
					} else {
						// just hover
						exsel_hover = sel;
					}
				} else {
					// we have an exhibit grabbed
					if(!act_grab) {
						// drop it
						Exhibit *ex = exsel_grab[i]->ex;
						exslot[i]->detach_exhibit();

						ExhibitSlot *slot = exman->nearest_empty_slot(vrhand[i].pos, 100);
						if(!slot) {
							debug_log("no empty slot nearby\n");
							if(ex->prev_slot && ex->prev_slot->empty()) {
								slot = ex->prev_slot;
								debug_log("using previous slot\n");
							}
						}

						if(slot) {
							ex->node->set_rotation(exslot[i]->node.get_rotation());
							slot->attach_exhibit(ex);
						} else {
							// nowhere to put it, stash it for later
							exman->stash_exhibit(ex);
							debug_log("no slots available, stashing\n");
						}

						*exsel_grab[i] = ExSelection::null;
					}
				}
			}
		}

		// if there are no grabs, and we're pointing with the right finger, override active
		if(!exsel_grab_left && !exsel_grab_right) {
			if(goatvr_action(1, GOATVR_ACTION_POINT)) {
				Ray ray;
				ray.origin = vrhand[1].pos;
				ray.dir = rotate(Vec3(0, 0, -1), vrhand[1].rot);
				//exsel_active = exman->select(ray);
				pointing = true;
			} else {
				exsel_active = ExSelection::null;
				pointing = false;
			}
		}

	} else {
		// check if an exhibit is hovered-over by mouse (only if we don't have one grabbed)
		if(!exsel_grab_mouse) {
			Ray ray = calc_pick_ray(prev_mx, prev_my);
			exsel_hover = exman->select(ray);
		}

		// TODO do this properly
		// set the position of the left hand at a suitable position for the exhibit UI
		dir = rotate(Vec3(-0.46, -0.1, -1), Vec3(0, 1, 0), deg_to_rad(-avatar.body_rot));
		exslot_left.node.set_position(avatar.pos + dir * 30); // magic: distance in front
	}

	if(!exslot_right.empty()) exslot_right.node.update(dt);
	// always update the left slot, because it's the anchor point of the exhibit ui
	exslot_left.node.update(dt);
}

void app_display()
{
	float dt = (float)(time_msec - prev_msec) / 1000.0f;
	prev_msec = time_msec;

	if(debug_gui) {
		ImGui::GetIOPtr()->DeltaTime = dt;
		ImGui::NewFrame();

		//ImGui::ShowTestWindow();
	}

	glClearColor(1, 1, 1, 1);

	if(opt.vr) {
		// VR mode
		goatvr_draw_start();
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		unsigned int gfbo = goatvr_get_fbo();

		update(dt);

		for(int i=0; i<2; i++) {
			// for each eye
			goatvr_draw_eye(i);
			if(gfbo) {
				vp_width = goatvr_get_fb_eye_width(i);
				vp_height = goatvr_get_fb_eye_height(i);

				// this is a lightweight operation
				goatvr_rtarg->create_wrap_fbo(gfbo, vp_width, vp_height);
				push_render_target(goatvr_rtarg, RT_FAKE);
			} else {
				vp_width = win_width / 2;
			}

			proj_matrix = goatvr_projection_matrix(i, NEAR_CLIP, FAR_CLIP);
			glMatrixMode(GL_PROJECTION);
			glLoadMatrixf(proj_matrix[0]);

			view_matrix = mouse_view_matrix * Mat4(goatvr_view_matrix(i));
			glMatrixMode(GL_MODELVIEW);
			glLoadMatrixf(view_matrix[0]);

			draw_scene();
			if(have_handtracking) {
				draw_vrhands();
			}

			if(debug_gui) {
				ImGui::Render();
			}

			if(gfbo) {
				pop_render_target(RT_FAKE);
			}
		}

		goatvr_draw_done();

		vp_width = win_width;
		vp_height = win_height;

		if(should_swap) {
			app_swap_buffers();
		}

	} else {
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		update(dt);

		proj_matrix.perspective(deg_to_rad(50.0), win_aspect, NEAR_CLIP, FAR_CLIP);
		glMatrixMode(GL_PROJECTION);
		glLoadMatrixf(proj_matrix[0]);

		view_matrix = mouse_view_matrix;
		glMatrixMode(GL_MODELVIEW);
		glLoadMatrixf(view_matrix[0]);

		draw_scene();

		if(!fb_srgb && sdr_post_gamma) {
			slow_post(sdr_post_gamma);
			glUseProgram(0);
		}

		if(debug_gui) {
			ImGui::Render();
		}
		app_swap_buffers();
	}
	assert(glGetError() == GL_NO_ERROR);

	calc_framerate();
}


static void draw_scene()
{
	rend->draw();
	exman->draw();

	if(have_handtracking) {
		glUseProgram(0);
		glPushAttrib(GL_ENABLE_BIT);
		glDisable(GL_LIGHTING);
		glBegin(GL_LINES);
		for(int i=0; i<2; i++) {
			if(vrhand[i].valid) {
				glColor3f(i, 1 - i, i);
			} else {
				glColor3f(0.5, 0.5, 0.5);
			}
			Vec3 v = vrhand[i].pos;
			Vec3 dir = rotate(Vec3(0, 0, -1), vrhand[i].rot) * 20.0f;
			Vec3 up = rotate(Vec3(0, 1, 0), vrhand[i].rot) * 10.0f;
			Vec3 right = rotate(Vec3(1, 0, 0), vrhand[i].rot) * 10.0f;

			if(i == 1 && pointing) {
				dir *= 1000.0f;
			}

			glVertex3f(v.x, v.y, v.z);
			glVertex3f(v.x + dir.x, v.y + dir.y, v.z + dir.z);
			glVertex3f(v.x - right.x, v.y - right.y, v.z - right.z);
			glVertex3f(v.x + right.x, v.y + right.y, v.z + right.z);
			glVertex3f(v.x - up.x, v.y - up.y, v.z - up.z);
			glVertex3f(v.x + up.x, v.y + up.y, v.z + up.z);
		}
		glEnd();
		glPopAttrib();
	}

	if(debug_gui && dbg_sel_node) {
		AABox bvol = dbg_sel_node->get_bounds();
		draw_geom_object(&bvol);
	}

	if(show_walk_mesh && mscn->walk_mesh) {
		glPushAttrib(GL_ENABLE_BIT);
		glEnable(GL_BLEND);
		glBlendFunc(GL_ONE, GL_ONE);
		glEnable(GL_POLYGON_OFFSET_FILL);

		glUseProgram(0);

		glPolygonOffset(-1, 1);
		glDepthMask(0);

		glColor3f(0.3, 0.08, 0.01);
		mscn->walk_mesh->draw();

		glDepthMask(1);

		glPopAttrib();
	}

	exui_draw();

	print_text(Vec2(9 * win_width / 10, 20), Vec3(1, 1, 0), "fps: %.1f", framerate);
	draw_ui();
}


void app_reshape(int x, int y)
{
	glViewport(0, 0, x, y);
	goatvr_set_fb_size(x, y, 1.0f);
	debug_gui_reshape(x, y);

	vp_width = x;
	vp_height = y;
}

void app_keyboard(int key, bool pressed)
{
	unsigned int mod = app_get_modifiers();

	if(debug_gui && !(pressed && (key == '`' || key == 27))) {
		debug_gui_key(key, pressed, mod);
		return;	// ignore all keystrokes when GUI is visible
	}

	if(pressed) {
		switch(key) {
		case 27:
			app_quit();
			break;

		case '\n':
		case '\r':
			if(mod & MOD_ALT) {
				app_toggle_fullscreen();
			}
			break;

		case '`':
			debug_gui = !debug_gui;
			show_message("debug gui %s", debug_gui ? "enabled" : "disabled");
			break;

		case 'm':
			app_toggle_grab_mouse();
			show_message("mouse %s", app_is_mouse_grabbed() ? "grabbed" : "released");
			break;

		case 'w':
			if(mod & MOD_CTRL) {
				show_walk_mesh = !show_walk_mesh;
				show_message("walk mesh: %s", show_walk_mesh ? "on" : "off");
			}
			break;

		case 'c':
			if(mod & MOD_CTRL) {
				noclip = !noclip;
				show_message(noclip ? "no clip" : "clip");
			}
			break;

		case 'f':
			toggle_flight();
			break;

		case 'p':
			if(mod & MOD_CTRL) {
				fb_srgb = !fb_srgb;
				show_message("gamma correction for non-sRGB framebuffers: %s\n", fb_srgb ? "off" : "on");
			}
			break;

		case '=':
			walk_speed *= 1.25;
			show_message("walk speed: %g", walk_speed);
			break;

		case '-':
			walk_speed *= 0.75;
			show_message("walk speed: %g", walk_speed);
			break;

		case ']':
			mouse_speed *= 1.2;
			show_message("mouse speed: %g", mouse_speed);
			break;

		case '[':
			mouse_speed *= 0.8;
			show_message("mouse speed: %g", mouse_speed);
			break;

		case 'b':
			show_blobs = !show_blobs;
			show_message("blobs: %s\n", show_blobs ? "on" : "off");
			break;

		case ' ':
			goatvr_recenter();
			show_message("VR recenter\n");
			break;

		case KEY_UP:
			exui_scroll(-1);
			break;

		case KEY_DOWN:
			exui_scroll(1);
			break;

		case KEY_LEFT:
			exui_change_tab(-1);
			break;

		case KEY_RIGHT:
			exui_change_tab(1);
			break;

		case KEY_F5:
		case KEY_F6:
		case KEY_F7:
		case KEY_F8:
			dbg_key_pending |= 1 << (key - KEY_F5);
			break;
		}
	}

	if(key < 256 && !(mod & (MOD_CTRL | MOD_ALT))) {
		keystate[key] = pressed;
	}
}

void app_mouse_button(int bn, bool pressed, int x, int y)
{
	static int press_x, press_y;

	if(debug_gui) {
		debug_gui_mbutton(bn, pressed, x, y);
		return;	// ignore mouse events while GUI is visible
	}

	prev_mx = x;
	prev_my = y;
	bnstate[bn] = pressed;

	if(bn == 0) {
		ExSelection sel;
		Ray ray = calc_pick_ray(x, y);
		sel = exman->select(ray);

		if(pressed) {
			if(sel && (app_get_modifiers() & MOD_CTRL)) {
				exsel_grab_mouse = sel;
				Vec3 pos = sel.ex->node->get_position();
				debug_log("grabbing... (%g %g %g)\n", pos.x, pos.y, pos.z);
				exslot_mouse.node.set_position(pos);
				exslot_mouse.node.set_rotation(sel.ex->node->get_rotation());
				exslot_mouse.attach_exhibit(sel.ex, EXSLOT_ATTACH_TRANSIENT);
				if(exsel_active) {
					exsel_active = ExSelection::null;	// cancel active on grab
				}
			}
			press_x = x;
			press_y = y;

		} else {
			if(exsel_grab_mouse) {
				// cancel grab on mouse release
				Exhibit *ex = exsel_grab_mouse.ex;
				Vec3 pos = exslot_mouse.node.get_position();

				debug_log("releasing at %g %g %g ...\n", pos.x, pos.y, pos.z);

				exslot_mouse.detach_exhibit();

				ExhibitSlot *slot = exman->nearest_empty_slot(pos, 100);
				if(!slot) {
					debug_log("no empty slot nearby\n");
					if(ex->prev_slot && ex->prev_slot->empty()) {
						slot = ex->prev_slot;
						debug_log("using previous slot\n");
					}
				}

				if(slot) {
					slot->attach_exhibit(ex);
				} else {
					// nowhere to put it, so stash it for later
					exslot_mouse.detach_exhibit();
					exman->stash_exhibit(ex);
					debug_log("no slots available, stashing\n");
				}

				exsel_grab_mouse = ExSelection::null;
			}

			if(abs(press_x - x) < 5 && abs(press_y - y) < 5) {
				exsel_active = sel;	// select or deselect active exhibit
				if(sel) {
					debug_log("selecting...\n");
				} else {
					debug_log("deselecting...\n");
				}
			}

			press_x = press_y = INT_MIN;
		}
	}
}

static inline void mouse_look(float dx, float dy)
{
	float scrsz = (float)win_height;
	avatar.body_rot += dx * 512.0 / scrsz;
	avatar.head_alt += dy * 512.0 / scrsz;

	if(avatar.head_alt < -90) avatar.head_alt = -90;
	if(avatar.head_alt > 90) avatar.head_alt = 90;
}

static void mouse_zoom(float dx, float dy)
{
	cam_dist += dy * 0.1;
	if(cam_dist < 0.0) cam_dist = 0.0;
}

void app_mouse_motion(int x, int y)
{
	if(debug_gui) {
		debug_gui_mmotion(x, y);
		return;	// ignore mouse events while GUI is visible
	}

	int dx = x - prev_mx;
	int dy = y - prev_my;
	prev_mx = x;
	prev_my = y;

	if(!dx && !dy) return;

	if(exsel_grab_mouse) {
		Vec3 pos = exslot_mouse.node.get_node_position();
		Vec3 dir = transpose(view_matrix.upper3x3()) * Vec3(dx * 1.0, dy * -1.0, 0);

		exslot_mouse.node.set_position(pos + dir);
	}

	if(bnstate[2]) {
		mouse_look(dx, dy);
	}
}

void app_mouse_delta(int dx, int dy)
{
	if(bnstate[2]) {
		mouse_zoom(dx * mouse_speed, dy * mouse_speed);
	} else {
		mouse_look(dx * mouse_speed, dy * mouse_speed);
	}
}

void app_mouse_wheel(int dir)
{
	if(debug_gui) {
		debug_gui_wheel(dir);
	}
}

void app_gamepad_axis(int axis, float val)
{
	switch(axis) {
	case 0:
		joy_move.x = val;
		break;
	case 1:
		joy_move.y = val;
		break;

	case 2:
		joy_look.x = val;
		break;
	case 3:
		joy_look.y = val;
		break;
	}
}

void app_gamepad_button(int bn, bool pressed)
{
	gpad_bnstate[bn] = pressed;

	if(pressed) {
		switch(bn) {
		case GPAD_LSTICK:
			toggle_flight();
			break;

		case GPAD_X:
			show_blobs = !show_blobs;
			show_message("blobs: %s\n", show_blobs ? "on" : "off");
			break;

		case GPAD_START:
			goatvr_recenter();
			show_message("VR recenter\n");
			break;

		default:
			break;
		}
	}
}

static void toggle_flight()
{
	static float prev_walk_speed = -1.0;
	if(prev_walk_speed < 0.0) {
		noclip = true;
		prev_walk_speed = walk_speed;
		walk_speed = 1000.0;
		show_message("fly mode\n");
	} else {
		noclip = false;
		walk_speed = prev_walk_speed;
		prev_walk_speed = -1.0;
		show_message("walk mode\n");
	}
}

static void calc_framerate()
{
	//static int ncalc;
	static int nframes;
	static long prev_upd;

	long elapsed = time_msec - prev_upd;
	if(elapsed >= 1000) {
		framerate = (float)nframes / (float)(elapsed * 0.001);
		nframes = 1;
		prev_upd = time_msec;

		/*if(++ncalc >= 5) {
			printf("fps: %f\n", framerate);
			ncalc = 0;
		}*/
	} else {
		++nframes;
	}
}

static Ray calc_pick_ray(int x, int y)
{
	float nx = (float)x / (float)win_width;
	float ny = (float)(win_height - y) / (float)win_height;

	last_pick_ray = mouse_pick_ray(nx, ny, view_matrix, proj_matrix);
	return last_pick_ray;
}
